if not InspectEquip then return end

local IE = InspectEquip
local IS = InspectEquip_ItemSources
local L = LibStub("AceLocale-3.0"):GetLocale("InspectEquip")

local band = bit.band
local bor = bit.bor
local tinsert = table.insert
local strfind = string.find

-- check for 5.0+ client, because EJ api was changed
local mop = select(4, GetBuildInfo()) >= 50000

-- Database
InspectEquipLocalDB = {}

-- GUI
local bar, barText
local coUpdate

function IE:InitLocalDatabase()
	if not InspectEquipLocalDB then InspectEquipLocalDB = {} end
	setmetatable(InspectEquipLocalDB, {__index = {
		Zones = {}, Bosses = {}, Items = {}
	}})

	local _, currentBuild = GetBuildInfo()

	-- create database if not present or outdated (or wrong locale)
	if (InspectEquipLocalDB.ClientBuild ~= currentBuild) or (InspectEquipLocalDB.Locale ~= GetLocale()) then
		self:CreateLocalDatabase()
	end
end

local function createUpdateGUI()

	bar = CreateFrame("STATUSBAR", nil, UIParent, "TextStatusBar")
	bar:SetWidth(300)
	bar:SetHeight(30)
	bar:SetPoint("CENTER", 0, -100)
	bar:SetBackdrop({
		bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = 1, tileSize = 10, edgeSize = 10,
		insets = {left = 1, right = 1, top = 1, bottom = 1}
    })
	bar:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar")
	bar:SetBackdropColor(0, 0, 0, 1)
    bar:SetStatusBarColor(0.6, 0.6, 0, 0.4)
    bar:SetMinMaxValues(0, 100)
	bar:SetValue(0)

	barText = bar:CreateFontString(nil, "OVERLAY", "GameFontHighlight")
    barText:SetPoint("CENTER", bar, "CENTER")
    barText:SetJustifyH("CENTER")
    barText:SetJustifyV("CENTER")
    barText:SetTextColor(1, 1, 1)
    barText:SetText("InspectEquip: " .. L["Updating database..."])

end

local ejDisabled = false
local ejToggle = ToggleEncounterJournal

-- this function disables the EJ GUI during update
-- because using the EJ during update will result in an invalid database
-- (other addons may still use the EJ api and possibly destroy the DB that way...)
local function DisableEJ()
	if ejDisabled then return end
	ejDisabled = true

	-- hide EJ
	if EncounterJournal then
		if EncounterJournal:IsShown() then
			EncounterJournal:Hide()
		end
	end

	ejToggle = ToggleEncounterJournal
	ToggleEncounterJournal = function() end
	EJMicroButton:Disable()
end

local function EnableEJ()
	if not ejDisabled then return end
	ejDisabled = false

	ToggleEncounterJournal = ejToggle
	EJMicroButton:Enable()
	EJMicroButton:SetAlpha(1)
end

local function EndUpdate()
	bar:SetScript("OnUpdate", nil)
	bar:Hide()
	EnableEJ()
	coUpdate = nil
end

local function UpdateTick()
	if coUpdate then
		local ok, msg = coroutine.resume(coUpdate)
		if not ok then
			EndUpdate()
			message("[InspectEquip] Could not update database: " .. msg)
		end
	end
end

local function UpdateBar()
	bar:SetValue(bar:GetValue() + 1)
	coroutine.yield()
end

local function GetReverseMapping(baseMap)
	local map = {}
	local id, name
	for id, name in pairs(baseMap) do
		map[name] = id
	end
	return map
end

local function DifficultyToMode(diff, raid)
	if diff == 1 then
		-- 5 normal / 10 normal
		return raid and 8 or 2
	elseif diff == 2 then
		-- 5 heroic / 25 normal
		return raid and 16 or 1
	elseif diff == 3 then
		-- 10 heroic
		return 32
	elseif diff == 4 then
		-- 25 heroic
		return 64
	elseif diff == 5 then
		-- lfr
		return 128
	end
end

local function AddToDB(tempDB, itemID, zoneID, bossID, mode)
	local sources = tempDB[itemID]
	local entry
	if sources then
		for _, entry in pairs(sources) do
			if (entry[1] == zoneID) and (entry[2] == bossID) then
				entry[3] = bor(entry[3], mode)
				return
			end
		end
		tinsert(sources, {zoneID, bossID, mode})
	else
		tempDB[itemID] = {{zoneID, bossID, mode}}
	end
end

local function SaveToDB(tempDB, entryType)
	local itemID, sources, entry
	for itemID, sources in pairs(tempDB) do
		local str = InspectEquipLocalDB.Items[itemID]
		local isEntry = IS.Items[itemID]

		-- loop through sources we found
		for _, entry in pairs(sources) do
			local entryStr = entryType .. "_" .. entry[1] .. "_" .. entry[3] .. "_" .. entry[2]

			-- skip if already in IS DB
			if not (isEntry and (strfind(";" .. isEntry .. ";", ";" .. entryStr .. ";"))) then
				if str then
					str = str .. ";" .. entryStr
				else
					str = entryStr
				end
			end

		end

		InspectEquipLocalDB.Items[itemID] = str
	end
end

local function GetInstanceCount(isRaid)
	local i = 1
	local id
	repeat
		id = EJ_GetInstanceByIndex(i, isRaid)
		i = i + 1
	until not id
	return i - 2
end

local function UpdateFunction()
	local _, i, j, tier

	-- reset EJ
	DisableEJ()
	EJ_ResetLootFilter()
	EJ_ClearSearch()

	-- init/reset database
	local db = InspectEquipLocalDB
	db.Zones = {}
	db.Bosses = {}
	db.Items = {}
	db.NextZoneID = 1000
	db.NextBossID = 1000

	-- count total number of instances
	-- as of 5.x there are multiple tiers (classic, bc, wotlk, cata, mop...)
	local tierCount = EJ_GetNumTiers()
	local insCount = 0
	for tier = 1, tierCount do
		local tierName = EJ_GetTierInfo(tier)
		EJ_SelectTier(tier)
		-- IE:Print("Tier: " .. tostring(tier) .. " = " .. tierName)
		local dungeonCount = GetInstanceCount(false)
		local raidCount = GetInstanceCount(true)
		-- IE:Print(" ==> " .. tostring(dungeonCount) .. " Dungeons, " .. tostring(raidCount) .. " Raids")
		insCount = insCount + dungeonCount + raidCount
	end

	-- set bar max value
	bar:SetMinMaxValues(0, insCount + 2)

	-- get IS mapping for zone/boss name -> zone/boss id
	local zoneMap = GetReverseMapping(IS.Zones)
	local bossMap = GetReverseMapping(IS.Bosses)

	UpdateBar()

	-- temp db to allow for merging of modes
	local tempDungeonDB = {}
	local tempRaidDB = {}

	local waitCycles = 0
	-- get loot
	for tier = 1, tierCount do
		EJ_SelectTier(tier)

		-- loop through all instances of this tier, dungeons first, then raids
		local isRaid = false
		while true do
			i = 1
			local insID, insName = EJ_GetInstanceByIndex(i, isRaid)
			local tempDB = isRaid and tempRaidDB or tempDungeonDB

			while insID do

				-- get zone id for db
				local zoneID = zoneMap[insName]
				if not zoneID then
					-- zone is not in db
					zoneID = db.NextZoneID
					db.Zones[zoneID] = insName
					zoneMap[insName] = zoneID
					db.NextZoneID = db.NextZoneID + 1
				end

				EJ_SelectInstance(insID)
				local diff
				local diffOffset = isRaid and 2 or 0
				-- EJ_IsValidInstanceDifficulty and EJ_SetDifficulty use different values
				-- unfortunately the constants are local variables in the EJ implementation
				-- this code works with 5.0.4
				for diff = 1, 5 do
					if EJ_IsValidInstanceDifficulty(diff + diffOffset) then
						EJ_SetDifficulty(diff)
						local mode = DifficultyToMode(diff, isRaid)
						local n = EJ_GetNumLoot()
						--IE:Print("T=" .. tostring(tier) .. " I=" .. insName .. " D=" .. diff .. " M=" .. mode .. " ! #loot = " .. n)

						for j = 1, n do
							-- get item info
							local itemName, _, _, _, itemID, itemLink, encID = EJ_GetLootInfoByIndex(j)

							-- wait until data has arrived
							while not itemID do
								coroutine.yield()
								waitCycles = waitCycles + 1
								itemName, _, _, _, itemID, itemLink, encID = EJ_GetLootInfoByIndex(j)
							end

							local encName = EJ_GetEncounterInfo(encID)

							-- get boss id for db
							local bossID = bossMap[encName]
							if not bossID then
								-- boss is not in db
								bossID = db.NextBossID
								db.Bosses[bossID] = encName
								bossMap[encName] = bossID
								db.NextBossID = db.NextBossID + 1
							end

							-- add item to db
							AddToDB(tempDB, itemID, zoneID, bossID, mode)
						end

					end
				end

				-- next instance
				UpdateBar()
				i = i + 1
				insID, insName = EJ_GetInstanceByIndex(i, isRaid)
			end

			if isRaid then
				-- done with dungeons + raids, next tier
				break
			else
				-- done with dungeons, now do it again for raids
				isRaid = true
			end
		end
	end

	-- save to db
	SaveToDB(tempDungeonDB, "d")
	SaveToDB(tempRaidDB, "r")
	local _, currentBuild = GetBuildInfo()
	db.ClientBuild = currentBuild
	db.Locale = GetLocale()
	UpdateBar()

	-- Done
	EndUpdate()
end

-- for 4.x clients
-- not 100% sure if this still works correctly
local function UpdateFunction4()

	-- init/reset database
	local db = InspectEquipLocalDB
	db.Zones = {}
	db.Bosses = {}
	db.Items = {}
	db.NextZoneID = 1000
	db.NextBossID = 1000

	-- get instance count
	local insID, insName, diff
	local i = 1
	local j
	repeat
		insID, insName = EJ_GetInstanceByIndex(i, true)
		i = i + 1
	until not insID
	local insCount = i - 2

	bar:SetMinMaxValues(0, insCount + 2)

	-- get IS mapping for zone/boss name -> zone/boss id
	local zoneMap = GetReverseMapping(IS.Zones)
	local bossMap = GetReverseMapping(IS.Bosses)

	UpdateBar()

	-- temp db to allow for merging of modes
	local tempDB = {}

	-- get loot!
	i = 1
	repeat
		insID, insName = EJ_GetInstanceByIndex(i, true)
		if not insID then break end

		-- get zone id for db
		local zoneID = zoneMap[insName]
		if not zoneID then
			-- zone is not in db
			zoneID = db.NextZoneID
			db.Zones[zoneID] = insName
			zoneMap[insName] = zoneID
			db.NextZoneID = db.NextZoneID + 1
		end

		EJ_SelectInstance(insID)
		for diff = 1, 5 do
			if EJ_IsValidInstanceDifficulty(diff) then
				EJ_SetDifficulty(diff)
				local mode = DifficultyToMode(diff, true)
				local n = EJ_GetNumLoot()
				for j = 1, n do
					-- get item info
					local itemName, _, _, _, itemID, itemLink, encID = EJ_GetLootInfoByIndex(j)
					local encName = EJ_GetEncounterInfo(encID)

					-- get boss id for db
					local bossID = bossMap[encName]
					if not bossID then
						-- boss is not in db
						bossID = db.NextBossID
						db.Bosses[bossID] = encName
						bossMap[encName] = bossID
						db.NextBossID = db.NextBossID + 1
					end

					-- add item to db
					AddToDB(tempDB, itemID, zoneID, bossID, mode)

				end
			end
		end

		UpdateBar()
		i = i + 1
	until not insID

	-- save to db
	SaveToDB(tempDB, "r")
	local _, currentBuild = GetBuildInfo()
	db.ClientBuild = currentBuild
	db.Locale = GetLocale()
	UpdateBar()

	-- Done
	EndUpdate()
	--IE:Print("Database updated!")
end

function IE:CreateLocalDatabase()

	if coUpdate then
		-- update already in progress
		return
	end

	-- load encounter journal
	if not IsAddOnLoaded("Blizzard_EncounterJournal") then
		local loaded, reason = LoadAddOn("Blizzard_EncounterJournal")
		if not loaded then
			message("[InspectEquip] Could not load encounter journal: " .. reason)
		end
	end

	-- show progress bar
	createUpdateGUI()
	bar:Show()

	-- start update
	if mop then
		coUpdate = coroutine.create(UpdateFunction)
	else
		coUpdate = coroutine.create(UpdateFunction4)
	end
	local ok, msg = coroutine.resume(coUpdate)
	if ok then
		bar:SetScript("OnUpdate", UpdateTick)
	else
		EndUpdate()
		message("[InspectEquip] Could not update database: " .. msg)
	end

end

